--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local addonName, NS = ...

local Lib = NS.FizzleLib

--------------------------- Scroll List -------------------------------------------------

local function DefaultButtonUpdateDisplay(self, table, index, data, savedvalue) -- self = parent = row [frame], index = table index#
	self.Index = index
	self.Selected:Hide()
	if self:GetParent().LastIndex and self:GetParent().LastIndex == self.Index then
		if self:GetParent().LastSelected then
			self:GetParent().LastSelected:Hide()
		end
		self:GetParent().LastSelected = self.Selected
		self:GetParent().LastIndex = self.Index
		self.Selected:Show()
	else
		self.Selected:Hide()
	end

end

local function OnEnter(self)
	local parent = self:GetParent()
	if not parent.Table[self.index].tooltip then return end
	GameTooltip:SetOwner(self, anchor or "ANCHOR_LEFT")
	GameTooltip:SetText(parent.Table[self.index].tooltip, nil, nil, nil, nil, true)

end

local function OnLeave(self)
	local parent = self:GetParent()
	if not self.index or not parent.Table[self.index].tooltip then return end
	GameTooltip:Hide()
end

local function ScrollUpdate(self, numItems, numToDisplay, buttonHeight) --(FauxScrollFrame_Update) requirements end here for my needs so far , button, smallWidth, bigWidth, highlightFrame, smallHighlightWidth, bigHighlightWidth, alwaysShowScrollBar)
	local scrollBar, scrollChildFrame, scrollUpButton, scrollDownButton = FauxScrollFrame_GetChildFrames(self)
	local showScrollBar
	if numItems > numToDisplay then --or alwaysShowScrollBar then
		scrollBar:Show()
		showScrollBar = 1
	else
		scrollBar:SetValue(0)
		scrollBar:Hide()
	end
	local scrollFrameHeight = 0
	local scrollChildHeight = 0
	if numItems > 0 then
		scrollFrameHeight = (numItems - numToDisplay) * buttonHeight
		scrollChildHeight = numItems * buttonHeight
		if scrollFrameHeight < 0 then
			scrollFrameHeight = 0
		end
		scrollChildFrame:Show()
	else
		scrollChildFrame:Hide()
	end
	local maxRange = (numItems - numToDisplay) * buttonHeight
	if maxRange < 0 then
		maxRange = 0
	end
	scrollBar:SetMinMaxValues(0, maxRange)
	scrollBar:SetValueStep(buttonHeight)
	scrollBar:SetStepsPerPage(numToDisplay-1)
	scrollChildFrame:SetHeight(scrollChildHeight)
	-- Arrow button handling
	if scrollBar:GetValue() == 0 then
		scrollUpButton:Disable()
	else
		scrollUpButton:Enable()
	end
	if scrollBar:GetValue() - scrollFrameHeight == 0 then
		scrollDownButton:Disable()
	else
		scrollDownButton:Enable()
	end
end

local function Update_List(self)
	local table = self.Table
	if not table then return end
	local numOptions = #table
	local offset = self.offset or 0
	local index, button
	local entries = #self.Buttons
	local include = 0
	for i=1, #table do
		index = offset + i
		if table[index] then -- and (not self.Excludes or not self.Excludes[table[index][self.Value]]) then
			include = include + 1
			button = self.Buttons[include]
			button:Show()
			button.index = index
			local text = button.Text or button.text
			if text then
				text:SetText(table[index][button:GetParent().Display])
			end
			self.ButtonUpdate(button, table, index, self.Data, self.Data[self.Index])
			if include == entries then break end -- only loop enough to full #buttons
		end
	end
	if include < entries then
		for i = include + 1, entries do
			self.Buttons[i]:Hide()
		end
	end
	self.ListHeight = self.Buttons[1]:GetHeight() * include
	ScrollUpdate(self, numOptions, entries, self.Buttons[1]:GetHeight())
end

local function OnVerticalScroll(self, offset, itemHeight)
	self.ScrollBar:SetValue(offset)
	self.offset = math.floor((offset / itemHeight) + 0.5)
	Update_List(self)
end

local function CreateScrollButton(parent, id, width, height, buttonlayout, buttontemplate)
	local button = CreateFrame(buttontemplate or "Button", "$parentScrollButton"..id, parent) -- might want to change for CheckButton
	button:SetSize(width, height)
	buttonlayout(button)
	button.id = id
	button:SetScript("OnEnter", OnEnter)
	button:SetScript("OnLeave", OnLeave)
	button:HookScript("OnClick", function(self) 
		local parent = self:GetParent()
		parent.Data[parent.Index] = parent.Table[self.index][parent.Value]
		parent.Key = self.index
		if parent.PostUpdate then
			parent:PostUpdate()--.Table[self.index])
		end
		Update_List(parent)
	end)
	return button
end

function Lib:CreateScrollList(name, parent, data, index, table, display, value, displayentries, buttonupdatedisplay, buttonlayout, buttontemplate)
	local size = 50
	local height = size/displayentries
	parent = parent or UIParent
	local f = CreateFrame("ScrollFrame", "$parent"..name, parent, "FauxScrollFrameTemplate")
	f.ButtonUpdate = buttonupdatedisplay or DefaultButtonUpdateDisplay
	f:SetSize(size, size)
	f.ScrollBar = FauxScrollFrame_GetChildFrames(f)
	f.Table = table
	f.Display = display
	f.Value = value
	f.Buttons = {}
	for i=1, displayentries do
		local b = CreateScrollButton(f, i, size, height, buttonlayout or Lib.CreateDefaultListButton, buttontemplate)
		f.Buttons[i] = b
		if i == 1 then
			b:SetPoint("TOPLEFT", f, "TOPLEFT")--, 5, -5)
		else
			b:SetPoint("TOPLEFT", f.Buttons[i-1], "BOTTOMLEFT")
		end
	end
	f:SetScript("OnVerticalScroll", function(self, offset)
		local height = self:GetHeight()/#self.Buttons
		OnVerticalScroll(self, offset, height)
	end)
	f:SetScript("OnSizeChanged", function(self)
		local width = self:GetWidth()
		local height = self:GetHeight()/#self.Buttons
		for i=1, #self.Buttons do
			self.Buttons[i]:SetSize(width, height)
		end
		self:GetScript("OnVerticalScroll")(self, 0)
	end)
	function f:GetRow(row)
		return self.Buttons[row]
	end
	function f:NumRows()
		return #self.Buttons
	end
	if not buttonlayout then
		function f:SetFont(font, size, r, g, b, alpha)
			local NewFont, NewSize = self.Buttons[1].Text:GetFont()
			NewFont = font or NewFont
			NewSize = size or NewSize
			for i=1, #self.Buttons do
				self.Buttons[i].Text:SetFont(NewFont, NewSize, "")
				if r then
					self.Buttons[i].Text:SetTextColor(r, g, b, alpha)
				end
			end
		end
	end
	Lib:AddLabelFuncs(f)
	Lib.InitData(f, data, index, Update_List)

	return f
end
