--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local AddonName, NS = ...

local L

function NS:SetLocale(text)
	L = text
end

-- Add Latest SettingsVersions to main addon when changed

-- All character settings will need to be updated if needed
-- Function returns: true to run addon updates process, true to open config
function NS:TransitionxPac() 
	-- NOTE: WOW_PROJECT_MAINLINE xpac changes are handled via NewSettings
	-- 	This function is primarily used for Classic transitions eg. vanilla-TBC/WoTLK or TBC/WoTLK etc.
	local Settings = NS:GetSettings()
	-- Classic/TBC cloning can go Classic -> TBC or TBC <- Classsic
		-- Trasnsition from Classic: MiniMapTrackingFrame to TBC: MiniMapTracking.
	if WOW_PROJECT_ID == WOW_PROJECT_MAINLINE then
		Settings.WOW_PROJECT_ID = WOW_PROJECT_ID
		return
	end
	-- Move to Wrath
	if not Settings.WOW_PROJECT_ID or (WOW_PROJECT_ID == NS.WOW_Project.WrathClassic and Settings.WOW_PROJECT_ID > WOW_PROJECT_MAINLINE and Settings.WOW_PROJECT_ID < NS.WOW_Project.WrathClassic) then
		for k, v in pairs(Settings.Profiles) do
			if v.UI.buttontrays[1].right[2].button == "MiniMapTracking" and MiniMapTracking then
				v.UI.buttontrays[1].right[2].button = "MiniMapTracking"
				v.UI.buttontrays[1].right[2].scale = 0.67
				v.UI.buttontrays[1].right[2].offsety = -1
			end
			v.UI.buttontrays[1].right[1].scale = 0.56 -- Game Time Frame
			v.UI.buttontrays[1].right[1].offsety = 1
			if v.UI.buttontrays[1].left[1].button == "LibDBIcon10_BugSack" then
				v.UI.buttontrays[1].left[1].offsetx = -8
			end
		end
	end
--[[ TBC is no more
	-- WOW_PROJECT_ID > Classic and Settings says Classic then move to TBC (or later if TBS Skipped
	if not Settings.WOW_PROJECT_ID or (WOW_PROJECT_ID > NS.WOW_Project.Classic and Settings.WOW_PROJECT_ID == NS.WOW_Project.Classic) then
		for k, v in pairs(Settings.Profiles) do
			if v.UI.buttontrays[1].right[2].button == "MiniMapTrackingFrame" and MiniMapTracking then
				v.UI.buttontrays[1].right[2].button = "MiniMapTracking"
			end
		end
	end
]]--
		-- Trasnsition from TBC: MiniMapTracking to Classic: MiniMapTrackingFrame.
	if not Settings.WOW_PROJECT_ID or (WOW_PROJECT_ID == NS.WOW_Project.Classic and Settings.WOW_PROJECT_ID ~= NS.WOW_Project.Classic) then
		for k, v in pairs(Settings.Profiles) do
			if v.UI.buttontrays[1].right[2].button == "MiniMapTrackingFrame" and MiniMapTracking then
				v.UI.buttontrays[1].right[2].button = "MiniMapTracking"
			end
		end

	end
	-- Next xPacs go here
	-- Could be Current Retail to New Retail
	-- or Current Classic to New Classic (with possible New Classic to Curent Classic cloning???)
	-- or ????
	Settings.WOW_PROJECT_ID = WOW_PROJECT_ID
end

NS.NewSettings = {}
-- NS.NewSettingsNotes = {}

NS.NewSettings[1] = function(lastglobalupdate, lastnewsetting, addonsinitialised, addonsassigned, reset)
	if addonsinitialised then return end
	NS:DefaultsAddonConfigInitialiseAll()
	return false, false -- no init, no open config
end

NS.NewSettings[2] = function(lastglobalupdate, lastnewsetting, addonsinitialised, addonsassigned, reset)
	if addonsinitialised then return end
	FAUXMAZZLE.Defaults.UI.hotspots[3].script = "if not TinyPad then return end\nSlashCmdList[\"TINYPADSLASH\"](\"\")"
	local Settings = NS:GetSettings("Profiles")
	for k, v in pairs(Settings) do
		v.Version = nil
		v.CHAT.editfont.font = nil -- move from fixed game font
		v.CHAT.chatfont.font = NS:GetOption("SharedMediaFont") -- to equivelent in LibSharedMedia.
		if v.ADDONS.Meters and v.ADDONS.meters then
			NS.Lib:CopyTable(v.ADDONS.meters, v.ADDONS.Meters)
			v.ADDONS.meters = nil
		end
	end
	return false, false -- no init, no open config
end

NS.NewSettings[3] = function(lastglobalupdate, lastnewsetting, addonsinitialised, addonsassigned, reset)
	if addonsassigned or reset then return end
	SetChatWindowUninteractable(4, false)
	local Settings = NS:GetSettings("Profiles")
	for k, v in pairs(Settings) do
		if v.ADDONS and v.ADDONS.TrinketMenu then
			table.wipe(v.ADDONS.TrinketMenu)
			v.ADDONS.TrinketMenu = nil
		end
		if v.ADDONS and v.ADDONS.Meters then
			v.ADDONS.Meters.configs = true
		end
	end
	local AddonName = "AutoBar"
	local Lookup = AddonName
	local yos = NS.Versions[AddonName][WOW_PROJECT_ID].mainyos
	if NS:IsValidVersion(NS.WOW_Project.Classic) then
		Lookup = Lookup.."Classic"
	end
	if IsAddOnLoaded(Lookup) then
		local Timer
		Timer = C_Timer.NewTicker(0.5, function()
			if not AutoBarClassBarBasicDriver then return end
			local ABasic = AutoBarClassBarBasicDriver
			Timer:Cancel()
			local Trink = NS:GetFrame("Utility", "Trinket2")
			local _, MainScale = NS.SetAutoBarScales()
			ABasic:SetScale(MainScale)
			ABasic:ClearAllPoints()
			if ABasic.FMSetPoint then
				ABasic:SetPoint("BOTTOMLEFT", Trink, "BOTTOMRIGHT", 0, yos)
			else
				ABasic:SetPoint("BOTTOMLEFT", Trink, "BOTTOMRIGHT", 0, yos)
			end
			local l = ABasic:GetLeft()
			local b = ABasic:GetBottom()
			local s = ABasic:GetEffectiveScale()
			if l and b then
				AutoBarDB.account.barList.AutoBarClassBarBasic.posX = l * s
				AutoBarDB.account.barList.AutoBarClassBarBasic.posY = b * s
				AutoBarDB.account.barList.AutoBarClassBarBasic.scale = MainScale
				LibStub("AceConfigRegistry-3.0"):NotifyChange(AddonName)
			end
		end)
	end
end

NS.NewSettings[4] = function(lastglobalupdate, lastnewsetting, addonsinitialised, addonsassigned, reset)
	if addonsassigned or reset then return end
	local AddonName = "AutoBar"
	local Configure
--	local Settings = NS:GetSettings().Defaults.Addons
--	if not Settings.AutoBar or not Settings.AutoBar.Config or Settings.AutoBar.Config < 4 then
--		UpdateAddonDefaultConfig("AutoBar", "Config", 4)
	if lastglobalupdate < 4 then
		Configure = true
	end
	if NS:IsValidVersion(NS.WOW_Project.Classic) then
		AddonName = AddonName.."Classic"
	end
	local loaded = IsAddOnLoaded(AddonName)
	if loaded then
		Settings = NS:GetSettings("Addons")
		Settings.AutoBar = Settings.AutoBar or {}
		Settings.AutoBar.inactive = nil
		if NS:IsValidVersion(NS.WOW_Project.Classic) then
	 		return
	 	end
	end
	if not NS:IsValidVersion(WOW_PROJECT_MAINLINE) then return end
	if loaded then
		local Timer
		Timer = C_Timer.NewTicker(0.5, function()
			if not AutoBarClassBarBasicDriver then return end
			Timer:Cancel()
			local ConfigFuncs = NS:GetConfigFuncs().addons["AutoBar"]
			local order, index = ConfigFuncs.order, ConfigFuncs.index
			ConfigFuncs = NS:GetConfigFuncs()[order][index]
			if Configure then
				ConfigFuncs.config()
			end
			ConfigFuncs.assign()
		end)
	end
	return false, Configure -- no init, open config if first time
end

NS.NewSettings[5] = function(lastglobalupdate, lastnewsetting, addonsinitialised, addonsassigned, reset)
	if addonsassigned or reset then return end
--	local RFAdded = NS:GetSettings("RaidFrames1")
	if NS:GetSettings().Version == 4 then
		NS:GetSettings().Version = 3 -- reset to last global
	end
	local ConfigFuncs = NS:GetConfigFuncs()
	local Config = ConfigFuncs.addons["RaidFrames1"] 
	local Assign = ConfigFuncs[Config.order][Config.index].assign
	Assign()
--	NS.ReasonText = L.GetText(L.NewSettingsText, L.RaidFrames, L.Installed)
	return false, true -- no init, open config
end

NS.NewSettings[6] = function(lastglobalupdate, lastnewsetting, addonsinitialised, addonsassigned, reset)
	if addonsassigned or reset then return end -- New characters will get this through the normal process
	-- 
	local function InitFor6()
		local ChannelList = {} -- Save any channels set (not going messages) before FCF_ResetChatWindows() clears them
		for i=5, 9 do -- 9-10 done as part of NS.DefaultsInitialiseChat
			local List = NS:ListChannelNames(GetChatWindowChannels(i))
			if List then
				ChannelList[i] = List
			end
		end
		NS.DefaultsInitialiseChat()
		for k, v in pairs(ChannelList) do  -- Restore saved channels
			local f = _G["ChatFrame"..k]
			for i=1, #v do
				if not NS:ChannelUpdateDelayed(k, "add", v[i], GetChannelList()) then
					ChatFrame_AddChannel(f, v[i])
				end
			end
		end
	end
	NS:GetOption("RegisterForInit")(nil, "PLAYER_ENTERING_WORLD", "NewSettings6", 1, InitFor6)

	local ConfigFuncs = NS:GetConfigFuncs().addons["Prat-3.0"] -- Modify dom Prat settings
	local order, index = ConfigFuncs.order, ConfigFuncs.index
	ConfigFuncs = NS:GetConfigFuncs()[order][index]
	ConfigFuncs.config()
	ConfigFuncs.assign()
--	NS.ReasonText = L.GetText(L.NewSettingsText, L.Chat, L.Options .. " " .. L.Reset)
	return false, true -- no init, open config only if the RFs weren't new this reest
end

NS.NewSettings[7] = function(lastglobalupdate, lastnewsetting, addonsinitialised, addonsassigned, reset)
	local Settings = NS:GetSettings("AddOns")
	if IsAddOnLoaded("HitsMode5") then
		DisableAddOn("HitsMode5", true)
	end
	if Settings.HitsMode5 then
		Settings.HitsMode5 = nil
	end
	if not Settings.CombatLog then
		Settings.CombatLog = {}
	end
	ChatFrame_RemoveAllMessageGroups(ChatFrame3)
	local ChatNames = {
		[1] = L.General,
		[2] = L.CombatLog,
		[3] = VOICE,
		[4] = L.Loot,
		[5] = L.Guild,
		[6] = L.Party,
		[7] = L.Whisper,
		[8] = L.Trade,
	}
	for i=2, #ChatNames do
		SetChatWindowName(i, ChatNames[i])
	end
	Settings = NS:GetSettings("Chat").channelbuttons
	local Nine, Ten
	for k, v in pairs(Settings) do
		if v.id == 9 then
			Nine = true
		end
		if v.id == 10 then
			Ten = true
		end
	end
	if not Nine then
		SetChatWindowShown(9, false)
		ChatFrame9:Hide()
	end
	if not Ten then
		SetChatWindowShown(10, false)
		ChatFrame10:Hide()
	end
	if not addonsassigned then -- this wasn't a new install for this character so fix for HitsMode if required
		local currentFilter = Blizzard_CombatLog_Filters.currentFilter
		if not Blizzard_CombatLog_Filters or Blizzard_CombatLog_Filters.filters[currentFilter].name ~= "FauxMazzle" then
			local ConfigFuncs = NS:GetConfigFuncs().addons["CombatLog"]
			local order, index = ConfigFuncs.order, ConfigFuncs.index
			ConfigFuncs = NS:GetConfigFuncs()[order][index]
			ConfigFuncs.config()
			NS.ReasonText = L.GetText(L.NewSettingsText, L.CombatLog, L.Reset)
			return false, true -- no init, open config
		end
	end
	return false, false -- no init, no open config
end

NS.NewSettings[8] = function(lastglobalupdate, lastnewsetting, addonsinitialised, addonsassigned, reset)
-- Dragonflight Beta (now overridden by [9] for anyone who installed the Beta on Live)
--[[
	local xpac = GetExpansionLevel()
	TocVersion = tonumber(select(4, GetBuildInfo()))
	if xpac > 8 or (xpac == 8 and TocVersion == 100000) then
		local Settings = NS:GetSettings()
		for k, v in pairs(Settings.Profiles) do
			v.UI.buttontrays[1].left[1] = { button={ "MinimapCluster", "MailFrame", }, system=true, scale=1.3, offsety=-1, } -- Mailbox Icon
			v.UI.buttontrays[1].right[1].scale=1.4 -- GameTimeFrame
			v.UI.buttontrays[1].right[1].offsetx=nil
			v.UI.buttontrays[1].right[1].offsety=nil
			v.UI.buttontrays[1].right[2] = { button={ "MinimapCluster", "Tracking", }, system=true, scale=1.3, }
			v.UI.buttontrays[1].right[3].button = "ExpansionLandingPageMinimapButton"
			v.UI.buttontrays[1].right[3].scale = 0.60
			for i, entry in ipairs(v.UI.buttontrays[1].left) do
				if entry.button == "QueueStatusMinimapButton" then
					entry.button = "Queuestatusbutton"
					break
				end
			end
		end
	end
]]--
	return false, false -- no init, no open config
end

NS.NewSettings[9] = function(lastglobalupdate, lastnewsetting, addonsinitialised, addonsassigned, reset)
-- Dragonflight Live (and for anyone who installed the Beta on Live) 
	if WOW_PROJECT_ID ~= WOW_PROJECT_MAINLINE then
		return false, false
	end
	local Settings = NS:GetSettings()
	for _, profile in pairs(Settings.Profiles) do
		if GetExpansionLevel() > 7 then
			for i, item in ipairs(profile.UI.buttontrays[1].left) do
				if item.button == "Queuestatusbutton" or item.button == "QueueStatusButton" or item.button == "QueueStatusMinimapButton" then
					item.button = "QueueStatusButton"
					item.scale = 0.56
					item.offsety=-1
				end
				if item.button == "MiniButton" then
					item.scale = 0.9
				end
				if item.button == "LibDBIcon10_BugSack" then
					item.scale=0.9
					item.offsety=nil
				end
			end
--			profile.UI.buttontrays[1].left[1] = { button={ "MinimapCluster", "MailFrame", }, system=true, scale=1.3, offsety=-1, } -- Mailbox Icon
			profile.UI.buttontrays[1].left[1] = NS.BaseMailEntry()
			profile.UI.buttontrays[1].left[1].scale=1.3
			profile.UI.buttontrays[1].right[1].scale=1.4 -- GameTimeFrame
			profile.UI.buttontrays[1].right[1].offsety=-4
			profile.UI.buttontrays[1].right[1].offsetx=nil
			profile.UI.buttontrays[1].right[2].button={ "MinimapCluster", "Tracking", }
			profile.UI.buttontrays[1].right[2].scale=1.3
			profile.UI.buttontrays[1].right[3].button="ExpansionLandingPageMinimapButton"
			profile.UI.buttontrays[1].right[3].scale=0.49
			profile.UI.buttontrays[1].right[3].offsetx=4
			profile.UI.buttontrays[1].right[3].offsety=0
		end
	end
	if IsAddOnLoaded("Dominos") and NS:GetSettings("Addons").Dominos then
		local ConfigFuncs = NS:GetConfigFuncs().addons["Dominos"]
		local order, index = ConfigFuncs.order, ConfigFuncs.index
		ConfigFuncs = NS:GetConfigFuncs()[order][index]
		ConfigFuncs.config(true)
		ConfigFuncs.assign(true)
		local profile = next(DominosDB.profiles)
		if profile == "FauxMazzle" then
			profile = next(DominosDB.profiles)
		end
		Dominos:SetProfile(profile)
		Dominos:SetProfile("FauxMazzle")
	end
--[[
	if IsAddOnLoaded("AutoBar") and NS:GetSettings("Addons").AutoBar then
		local Timer
		Timer = C_Timer.NewTicker(0.5, function()
			if not AutoBarClassBarBasicDriver then return end
			Timer:Cancel()
			local ConfigFuncs = NS:GetConfigFuncs().addons["AutoBar"]
			local order, index = ConfigFuncs.order, ConfigFuncs.index
			ConfigFuncs = NS:GetConfigFuncs()[order][index]
			ConfigFuncs.config()
			ConfigFuncs.assign()
		end)
	end
]]--
	if IsAddOnLoaded("DiscordUnitFrames") and NS:GetSettings("Addons").DiscordUnitFrames then
		local Timer
		Timer = C_Timer.NewTicker(0.5, function()
			if not DUF_PLAYER_INDEX then return end
			Timer:Cancel()
			local ConfigFuncs = NS:GetConfigFuncs().addons["DiscordUnitFrames"]
			local order, index = ConfigFuncs.order, ConfigFuncs.index
			ConfigFuncs = NS:GetConfigFuncs()[order][index]
			ConfigFuncs.config(true)
			ConfigFuncs.assign(true)
		end)
	end
--	NS.ReasonText = L.DragonFlightReset
	return false, true -- no init, open config
end
NS.NewSettings[10] = function(lastglobalupdate, lastnewsetting, addonsinitialised, addonsassigned, reset)
	if MiniMapLFGFrame and WOW_PROJECT_ID == NS.WOW_Project.WrathClassic  then
		NS.ReasonText = L.GetText(L.NewSettingsText, L.QueueStatus, L.Reset)
		local Settings = NS:GetSettings()
		for _, profile in pairs(Settings.Profiles) do
			if profile.UI.buttontrays[1].left[2] and profile.UI.buttontrays[1].left[2].button == "LibDBIcon10_BugSack" then
				profile.UI.buttontrays[1].left[2].offsetx = -4
			end
			if not profile.UI.buttontrays[1].left[2] or profile.UI.buttontrays[1].left[2].button ~= "MiniMapLFGFrame" then
				tinsert(profile.UI.buttontrays[1].left, 2, 
				{
					offsety = -1,
					scale = 0.64,
					offsetx = -6,
					button = "MiniMapLFGFrame",
					system = true,
				})
			end	
		end
		return false, true -- no init, open config
	end
	return false, false -- no init, no open config
end
NS.NewSettings[11] = function(lastglobalupdate, lastnewsetting, addonsinitialised, addonsassigned, reset)
	local Settings = NS:GetSettings("Profiles")
	for k, v in pairs(Settings) do -- Check all characters in case some haven't been used since the
		local disableCount = 0 -- "reason" return from C_AddOns.GetAddOnInfo(addon) changed to always a string
		for _, addon in pairs(v.ADDONS) do
			if addon.inactive == true then
				disableCount = disableCount + 1
			end
		end
		if disableCount > 8 then -- If character has disabled more the 8 addons then assume broken
			for x, reset in pairs(v.ADDONS) do
				if reset.inactive ~= nil then
					reset.inactive = false
				end
			end
		end
	end
	return false, false -- no init, no open config
end